/*
 * Copyright (C) 2002-2007 The Warp Rogue Team
 * Part of the Warp Rogue Project
 *
 * This software is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License.
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.
 *
 * See the license.txt file for more details.
 */

/*
 * Module Name: Area
 * Description: -
 */

#define Uses_Util
#define Uses_Sector
#define Uses_Event
#define Uses_Character
#define Uses_Object
#define Uses_Terrain
#define Uses_ProgramManager
#define Uses_Faction
#define Uses_Perception
#include "mheader.h"

#include "area.h"


#define	AREA_HEIGHT		64
#define	AREA_WIDTH		128

#define at(y, x) 		((y) * AREA_WIDTH + (x))




/*
 * the currently active (i.e. loaded) area
 */
static AREA *			ActiveArea = NULL;


/*
 * the NIL point
 */
static const AREA_POINT		AreaPointNil = {
	
	AREA_COORD_NIL,
	AREA_COORD_NIL
	
};


/*
 * area bounds
 */
static const AREA_SECTION	AreaBounds = {
	
	0, AREA_HEIGHT -1, 
	0, AREA_WIDTH -1
	
};




/*
 * area module clean up
 */
void area_clean_up(void)
{

	if (ActiveArea != NULL) {
		
		area_destroy();
	}
}



/*
 * allocates memory for an area
 */
AREA * area_new(void)
{
	AREA *area;
	AREA_POINT p;
	const SECTOR *default_sector;

	area = checked_malloc(sizeof *area);

	area->characters = list_new();
	area->objects = list_new();
	area->terrain = list_new();

	area->events = list_new();

	area->sector = checked_malloc(AREA_HEIGHT * AREA_WIDTH *
		sizeof *area->sector
	);

	default_sector = sector_template();

	for (p.y = 0; p.y < AREA_HEIGHT; p.y++) {
	for (p.x = 0; p.x < AREA_WIDTH; p.x++) {
		SECTOR *sector;

		sector = &area->sector[at(p.y, p.x)];

		*sector = *default_sector;
	}
	}

	return area;
}



/*
 * destroys the active area
 */
void area_destroy(void)
{
	/*
	 * this function is optimized for speed - nasty hacks!
	 */
	CHARACTER *character;
	LIST_NODE *node;

	list_destroy_with(ActiveArea->terrain, free);
	list_destroy_with(ActiveArea->objects, free);

	for (node = ActiveArea->characters->head;
		node != NULL;
		node = node->next) {

		character = (CHARACTER *)node->data;

		if (is_player_controlled_character(character)) {

			set_player_controlled_character(NULL);
		}

		if (character == player_character()) {

			set_player_character(NULL);
		}

		list_destroy(character->inventory);

		list_destroy(character->noticed_enemies);

		free(character);
	}

	list_destroy(ActiveArea->characters);

	list_destroy_with(ActiveArea->events, event_destroy);

	free(ActiveArea->sector);

	free(ActiveArea);

	ActiveArea = NULL;
}



/*
 * sets the active area
 */
void set_active_area(AREA *area)
{

	ActiveArea = area;
}



/*
 * returns the active area
 */
AREA * active_area(void)
{

	return ActiveArea;
}



/*
 * returns the terrain list of the active area
 */
LIST * area_terrain_list(void)
{

	return ActiveArea->terrain;
}



/*
 * returns the object list of the active area
 */
LIST * area_object_list(void)
{

	return ActiveArea->objects;
}



/*
 * returns the character list of the active area
 */
LIST * area_character_list(void)
{

	return ActiveArea->characters;
}



/*
 * returns the event list of the active area
 */
LIST * area_event_list(void)
{

	return ActiveArea->events;
}



/*
 * returns the number of characters in the active area
 */
N_CHARACTERS area_n_characters(void)
{

	return ActiveArea->characters->n_nodes;
}



/*
 * returns true if the coordinates of both area points are equal
 */
bool area_points_equal(const AREA_POINT *p1, const AREA_POINT *p2)
{

	if (p1->x == p2->x && p1->y == p2->y) {
		
		return true;
	}

	return false;
}



/*
 * moves an area point one step in the passed direction
 */
void move_area_point(AREA_POINT *area_point, DIRECTION direction)
{

	area_point->x += direction_modifier(COORD_X, direction);
	area_point->y += direction_modifier(COORD_Y, direction);	
}



/*
 * returns the estimated euclidian distance between two area points
 */
AREA_DISTANCE area_distance(const AREA_POINT *p1, const AREA_POINT *p2)
{
	AREA_DISTANCE xdiff, ydiff, distance;

	xdiff = abs(p1->x - p2->x);
	ydiff = abs(p1->y - p2->y);
	
	distance = xdiff + ydiff + max(xdiff, ydiff);
	distance >>= 1;

	return distance;
}



/*
 * returns the area bounds
 */
const AREA_SECTION * area_bounds(void)
{
	
	return &AreaBounds;
}



/*
 * returns true if the passed point is out of bounds
 */
bool out_of_area_bounds(const AREA_POINT *location)
{

	if (location->y < 0 ||
		location->x < 0 ||
		location->y >= AREA_HEIGHT ||
		location->x >= AREA_WIDTH) {

		return true;
	}

	return false;
}



/*
 * return the area point NIL
 */
const AREA_POINT * area_point_nil(void)
{

	return &AreaPointNil;
}



/*
 * places a terrain at the passed point
 */
void place_terrain(TERRAIN *terrain, const AREA_POINT *target_point)
{

	sector_at(target_point)->terrain = terrain;

	if (terrain->list_node != NULL) return;

	terrain->list_node = list_add(ActiveArea->terrain, terrain);
}



/*
 * removes a terrain object from the currently active area
 */
void remove_terrain(TERRAIN *terrain, LIST *terrain_events)
{

	remove_terrain_events(terrain, terrain_events);

	list_remove_node(ActiveArea->terrain, terrain->list_node);
	terrain->list_node = NULL;
}



/*
 * places an object at the passed point
 */
void place_object(OBJECT *object, const AREA_POINT *target_point)
{

	sector_at(target_point)->object = object;

	if (object->list_node != NULL) return;

	object->list_node = list_add(ActiveArea->objects, object);
}



/*
 * removes an object from the currently active area
 */
void remove_object(OBJECT *object, LIST *object_events)
{

	remove_object_events(object, object_events);

	list_remove_node(ActiveArea->objects, object->list_node);
	object->list_node = NULL;
}



/*
 * places a character at the passed point
 */
void place_character(CHARACTER *character, const AREA_POINT *target_point)
{
	LIST_NODE *node;

	sector_at(target_point)->character = character;
	character->location = *target_point;

	if (character->list_node != NULL) return;

	character->list_node = list_add(ActiveArea->characters,
		character
	);

	for (node = character->inventory->head;
		node != NULL;
		node = node->next) {
		OBJECT *object;

		object = (OBJECT *)node->data;

		object->list_node = list_add(ActiveArea->objects,
			object
		);
	}
}



/*
 * removes a character from the currently active area
 */
void remove_character(CHARACTER *character, LIST *character_events)
{
	LIST_NODE *node;

	remove_character_events(character, character_events);

	for (node = character->inventory->head;
		node != NULL;
		node = node->next) {
		OBJECT *object;

		object = (OBJECT *)node->data;

		remove_object(object, character_events);
	}

	list_remove_node(ActiveArea->characters,
		character->list_node
	);
	character->list_node = NULL;

	sector_at(&character->location)->character = NULL;
	character->location = *area_point_nil();
}



/*
 * returns the sector located at the passed point
 */
SECTOR * sector_at(const AREA_POINT *location)
{

	return &ActiveArea->sector[at(location->y, location->x)];
}



/*
 * returns the terrain located at the passed point
 */
TERRAIN * terrain_at(const AREA_POINT *location)
{

	return ActiveArea->sector[at(location->y, location->x)].terrain;
}



/*
 * returns the object located at the passed point
 */
OBJECT * object_at(const AREA_POINT *location)
{

	return ActiveArea->sector[at(location->y, location->x)].object;
}



/*
 * returns the character located at the passed point
 */
CHARACTER * character_at(const AREA_POINT *location)
{

	return ActiveArea->sector[at(location->y, location->x)].character;
}





