/*
 * Copyright (C) 2002-2007 The Warp Rogue Team
 * Part of the Warp Rogue Project
 *
 * This software is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License.
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.
 *
 * See the license.txt file for more details.
 */

/*
 * Module Name: DataFile
 * Description: -
 */

#define Uses_Ui
#define Uses_ProgramManager
#define Uses_Util
#define Uses_Equipment
#define Uses_Career
#define Uses_Terrain
#define Uses_Object
#define Uses_Stats
#define Uses_Faction
#define Uses_Perks
#define Uses_Inventory
#define Uses_Npc
#define Uses_Character
#define Uses_Stats
#define Uses_Psychic
#define Uses_Options
#define Uses_World
#define Uses_Dungeon
#define Uses_Perception
#include "mheader.h"

#include "rdb.h"

#include "datafile.h"



/*
 * config files
 */
#define KEYBINDINGS_FILE		"keys.rdb"
#define VIDEO_CONFIG_FILE		"video.rdb"
#define OPTIONS_FILE			"options.rdb"

/*
 * career files
 */
#define FILE_BASIC_CAREERS		"b_career.rdb"
#define FILE_ADVANCED_CAREERS		"a_career.rdb"

/*
 * perk files
 */
#define FILE_PERKS			"perks.rdb"

/*
 * NPC files
 */
#define MAX_NPC_FILES			5
#define FILE_IMPERIUM_NPCS		"imperium.rdb"
#define FILE_REBEL_NPCS			"rebels.rdb"
#define FILE_CHAOS_NPCS			"chaos.rdb"
#define FILE_ALIEN_NPCS			"aliens.rdb"
#define FILE_BEAST_NPCS			"beasts.rdb"

/*
 * object files
 */
#define MAX_OBJECT_FILES		7
#define FILE_CCW_OBJECTS		"ccw.rdb"
#define FILE_RCW_OBJECTS		"rcw.rdb"
#define FILE_DRUG_OBJECTS		"drugs.rdb"
#define FILE_ARMOUR_OBJECTS		"armours.rdb"
#define FILE_ENVIRONMENT_OBJECTS	"env.rdb"
#define FILE_MISC_OBJECTS		"misc.rdb"
#define FILE_PSYCHIC_BOLTS		"psybolts.rdb"

/*
 * terrain files
 */
#define FILE_TERRAIN			"terrain.rdb"

/*
 * world files
 */
#define FILE_WORLD			"world.rdb"




static char *		data_file_entry_description(char *, const char *);

static void		load_keybindings(void);
static void		load_options(void);

static void		load_terrain_files(void);
static void		load_terrain_file(void);

static void		load_object_files(void);
static void		load_object_file(FILE_INDEX);

static void		load_career_files(void);
static void		load_career_file(void);

static void		load_npc_files(void);
static void		load_npc_file(FILE_INDEX);
static void		load_npc_pp(CHARACTER *);

static void		load_world_file(void);

static void		load_dungeon_files(void);
static void		load_dungeon_file(const char *, const WORLD_POINT *);

static void		read_symbol_field(SYMBOL *);
static void		read_dynamic_color_field(DYNAMIC_COLOR *);
static void		read_stat_field(CHARACTER *, const char *);
static void		read_firing_mode_field(OBJECT_DATA *);
static void		read_dice_roll_field(DICE_ROLL *);
static OBJECT *		read_object_field(void);
static void		read_career_field(CHARACTER *);
static void		read_description_field(char *, const char *);
static void		read_key_field(KEY_CODE *);
static void		read_option_value_field(OPTION_VALUE *);
static void		read_boolean_field(bool *);
static void		read_ai_state_field(CHARACTER *);
static void		read_gender_field(GENDER *);
static void		read_name_field(char *, unsigned int);
static void		read_career_exit_field(CAREER *, CAREER_EXIT_INDEX);

static void		read_map_tile(void);
static void		read_map_field(WORLD_POINT *);



/*
 * object files
 */
static const char *	ObjectFile[MAX_OBJECT_FILES] = {

	FILE_CCW_OBJECTS,
	FILE_RCW_OBJECTS,
	FILE_ARMOUR_OBJECTS,
	FILE_DRUG_OBJECTS,
	FILE_ENVIRONMENT_OBJECTS,
	FILE_PSYCHIC_BOLTS,
	FILE_MISC_OBJECTS

};



/*
 * NPC files
 */
static const char *	NpcFile[MAX_NPC_FILES] = {

	FILE_IMPERIUM_NPCS,
	FILE_REBEL_NPCS,
	FILE_CHAOS_NPCS,
	FILE_ALIEN_NPCS,
	FILE_BEAST_NPCS

};



/*
 * the path of the currently processed file
 */
static char		DataPath[FILE_PATH_SIZE];



/*
 * sets the data path
 */
void set_data_path(PROGRAM_DIRECTORY directory, const char *file_name)
{

	get_file_path(DataPath, directory, file_name);
}



/*
 * returns the data path
 */
const char * data_path(void)
{

	return DataPath;
}



/*
 * loads the data files
 */
void load_data_files(void)
{

	load_keybindings();

	load_options();

	load_terrain_files();

	load_object_files();

	load_career_files();

	load_npc_files();

	load_world_file();
	
	load_dungeon_files();
}



/*
 * returns the description of a terrain
 */
char * data_file_terrain_description(char *description, const TERRAIN *terrain)
{

	set_data_path(DIR_TERRAIN, FILE_TERRAIN);

	data_file_entry_description(description,
		terrain_static_data(terrain)->name
	);

	return description;
}



/*
 * returns the description of an object
 */
char * data_file_object_description(char *description, const OBJECT *object)
{
	const OBJECT_DATA *object_data;

	clear_string(description);

	object_data = object_static_data(object);

	if (object_data->source_file == FILE_INDEX_NIL) {

		return description;
	}

	set_data_path(DIR_OBJECTS, ObjectFile[object_data->source_file]);

	data_file_entry_description(description, object_data->name);

	return description;
}



/*
 * returns the description of a career
 */
char * data_file_career_description(char *description, const CAREER *career)
{

	if (career->type == CAREER_TYPE_BASIC) {
		
		set_data_path(DIR_CHARACTERS, FILE_BASIC_CAREERS);

	} else if (career->type == CAREER_TYPE_ADVANCED) {
		
		set_data_path(DIR_CHARACTERS, FILE_ADVANCED_CAREERS);

	} else {
		
		die("*** CORE ERROR *** invalid career description call");
	}

	data_file_entry_description(description, career->name);

	return description;
}



/*
 * returns the description of a perk
 */
char * data_file_perk_description(char *description, PERK perk)
{
	clear_string(description);

	set_data_path(DIR_CHARACTERS, FILE_PERKS);

	data_file_entry_description(description, perk_name(perk));

	return description;
}



/*
 * returns the description of a character
 */
char * data_file_character_description(char *description,
	const CHARACTER *character
)
{

	clear_string(description);

	if (character->source_file != FILE_INDEX_NIL) {

		set_data_path(DIR_CHARACTERS, 
			NpcFile[character->source_file]
		);

		data_file_entry_description(description, character->name);
	}

	if (is_empty_string(description) &&
		character->career != CAREER_NIL) {
		CAREER *career;

		career = get_career_pointer(character->career);

		career_description(description, career);
	}

	return description;
}



/*
 * returns the description of a data file entry
 */
static char * data_file_entry_description(char *description,
	const char *entry_name
)
{
	bool entry_reached = false;

	rdb_open(data_path(), RDB_READ);

	clear_string(description);

	while (rdb_next_field()) {
		const char *field_name;

                field_name = rdb_field_name();

		if (strings_equal(field_name, "NAME")) {

			if (entry_reached) break;

			if (strings_equal(rdb_data_token(RDB_NEXT_TOKEN),
				entry_name)) {

				entry_reached = true;
			}
			continue;
		}

		if (entry_reached && 
			strings_equal(field_name, "DESCRIPTION")) {

			read_description_field(description, entry_name);
		}
	}

	rdb_close();

	return description;
}



/*
 * loads the keybindings
 */
static void load_keybindings(void)
{
	COMMAND command = CM_NIL;

	set_data_path(DIR_CONFIG, KEYBINDINGS_FILE);

	rdb_open(data_path(), RDB_READ);

        while (rdb_next_field()) {
		const char *field_name;

                field_name = rdb_field_name();

		if (strings_equal(field_name, "COMMAND")) {

			command = name_to_command(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (strings_equal(field_name, "KEY")) {
			KEY_CODE key;

			read_key_field(&key);

			set_command_key(command, key);

		} else {

			die("*** CORE ERROR *** invalid field " \
				"name: %s (%s)",
				field_name,
				data_path()
			);
		}

	}

	rdb_close();

}



/*
 * loads the options file
 */
static void load_options(void)
{

	set_data_path(DIR_CONFIG, OPTIONS_FILE);

	rdb_open(data_path(), RDB_READ);

        while (rdb_next_field()) {
		const char *field_name;
		OPTION_VALUE option_value;

                field_name = rdb_field_name();

		read_option_value_field(&option_value);

		if (strings_equal(field_name, "FULLSCREEN")) {

			option_set_value(OPT_FULLSCREEN,
				option_value
			);

		} else if (strings_equal(field_name, "BRIGHTNESS")) {

			option_set_value(OPT_BRIGHTNESS,
				option_value
			);

		} else if (strings_equal(field_name, "CENTER_ON_PLAYER")) {

			option_set_value(OPT_CENTER_ON_PLAYER,
				option_value
			);

		} else {

			die("*** CORE ERROR *** invalid field name: " \
				"%s (%s)",
				field_name,
				data_path()
			);
		}
	}

	rdb_close();
}



/*
 * loads the terrain files
 */
static void load_terrain_files(void)
{

	set_data_path(DIR_TERRAIN, FILE_TERRAIN);

	load_terrain_file();

	terrain_box_optimize();
}



/*
 * loads a terrain file
 */
static void load_terrain_file(void)
{
	char description[DESCRIPTION_SIZE];
	TERRAIN_DATA *terrain_data = NULL;

	rdb_open(data_path(), RDB_READ);

        while (rdb_next_field()) {
		const char *field_name;

                field_name = rdb_field_name();

		if (strings_equal(field_name, "NAME")) {

			terrain_data = terrain_box_new_terrain();

			clear_string(description);

			read_name_field(terrain_data->name,
				TERRAIN_NAME_SIZE
			);

		} else if (strings_equal(field_name, "SYMBOL")) {

			read_symbol_field(&terrain_data->symbol);

		} else if (strings_equal(field_name, "COLOR")) {

			read_dynamic_color_field(&terrain_data->color);

		} else if (strings_equal(field_name, "ATTRIBUTE")) {
			TERRAIN_ATTRIBUTE attribute;

			attribute = name_to_terrain_attribute(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

			terrain_data->attribute[attribute] = true;

		} else if (strings_equal(field_name, "DESCRIPTION")) {

			read_description_field(description, 
				terrain_data->name
			);

		} else {

			die("*** CORE ERROR *** invalid field name: " \
				"%s (%s)",
				field_name,
				data_path()
			);
		}


        }

	rdb_close();
}



/*
 * loads the object files
 */
static void load_object_files(void)
{
	FILE_COUNTER i;

	for (i = 0; i < MAX_OBJECT_FILES; i++) {

		set_data_path(DIR_OBJECTS, ObjectFile[i]);

		load_object_file(i);
	}

	object_box_optimize();
}



/*
 * loads an object file
 */
static void load_object_file(FILE_INDEX file_index)
{
	char description[DESCRIPTION_SIZE];
	OBJECT_DATA *object_data = NULL;

	rdb_open(data_path(), RDB_READ);

        while (rdb_next_field()) {
		const char *field_name;

                field_name = rdb_field_name();

		if (strings_equal(field_name, "NAME")) {

			object_data = object_box_new_object();

			clear_string(description);

			read_name_field(object_data->name,
				OBJECT_NAME_SIZE
			);

			object_data->source_file = file_index;

		} else if (strings_equal(field_name, "SYMBOL")) {

			read_symbol_field(&object_data->symbol);

		} else if (strings_equal(field_name, "COLOR")) {

			read_dynamic_color_field(&object_data->color);

		} else if (strings_equal(field_name, "TYPE")) {

			object_data->type = name_to_object_type(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (strings_equal(field_name, "SUBTYPE")) {

			object_data->subtype = name_to_object_subtype(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (strings_equal(field_name, "CONDITION")) {

			object_data->condition_max = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (strings_equal(field_name, "VALUE")) {

			object_data->value = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (strings_equal(field_name, "BUYABLE")) {

			read_boolean_field(&object_data->buyable);

		} else if (strings_equal(field_name, "WEIGHT")) {

			object_data->weight = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (strings_equal(field_name, "DAMAGE")) {

			read_dice_roll_field(&object_data->damage);

		} else if (strings_equal(field_name, "ARMOUR_VALUE")) {

			object_data->armour_value = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (strings_equal(field_name, "ACCURACY")) {

			object_data->firing_mode.accuracy = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (strings_equal(field_name, "SHOTS")) {

			object_data->charge_max = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (strings_equal(field_name, "FIRING_MODE")) {

			read_firing_mode_field(object_data);

		} else if (strings_equal(field_name, "RELOAD")) {

			object_data->reload = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (strings_equal(field_name, "ATTRIBUTE")) {
			OBJECT_ATTRIBUTE attribute;

			attribute = name_to_object_attribute(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

			object_data->attribute[attribute] = true;

		} else if (strings_equal(field_name, "REACH")) {

			object_data->reach = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (strings_equal(field_name, "PARRY_PENALTY")) {

			object_data->parry_penalty = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (strings_equal(field_name, "DESCRIPTION")) {

			read_description_field(description,
				object_data->name
			);

		} else {

			die("*** CORE ERROR *** invalid field name: %s (%s)",
				field_name,
				data_path()
			);
		}


        }

	rdb_close();
}



/*
 * loads the career files
 */
static void load_career_files(void)
{

	set_data_path(DIR_CHARACTERS, FILE_BASIC_CAREERS);
	load_career_file();

	set_data_path(DIR_CHARACTERS, FILE_ADVANCED_CAREERS);
	load_career_file();

	career_box_optimize();

#if defined(DEBUG)
	career_box_validate();
#endif

}



/*
 * loads a career file
 */
static void load_career_file(void)
{
	char description[DESCRIPTION_SIZE];
	CAREER *career = NULL;
	CAREER_EXIT_INDEX career_exit_index = 0;

	rdb_open(data_path(), RDB_READ);

        while (rdb_next_field()) {
		const char *field_name;

                field_name = rdb_field_name();

		if (strings_equal(field_name, "NAME")) {

			career = career_box_new_career();

			career_exit_index = 0;

			clear_string(description);

			read_name_field(career->name,
				CAREER_NAME_SIZE
			);

		} else if (strings_equal(field_name, "TYPE")) {

			career->type = name_to_career_type(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (strings_equal(field_name, "COLOR")) {

			career->color = name_to_color(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (is_short_stat_name(field_name)) {
			STAT i;

			i = name_to_stat(field_name, true);

			career->max_stat_advance[i] = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (strings_equal(field_name, "PERK")) {
			PERK perk;

			perk = name_to_perk(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

			career->perk[perk] = true;


		} else if (strings_equal(field_name, "OBJECT")) {
			OBJECT *object;

			object = read_object_field();

			list_add(career->trappings, object);

		} else if (strings_equal(field_name, "EXIT")) {
			
			read_career_exit_field(career, career_exit_index);

			++career_exit_index;

		} else if (strings_equal(field_name, "DESCRIPTION")) {

			read_description_field(description, career->name);

		} else {

			die("*** CORE ERROR *** invalid field name: %s (%s)",
				field_name,
				data_path()
			);
		}
        }

        rdb_close();
}



/*
 * loads the NPC files
 */
static void load_npc_files(void)
{
	FILE_COUNTER i;

	for (i = 0; i < MAX_NPC_FILES; i++) {

		set_data_path(DIR_CHARACTERS, NpcFile[i]);

		load_npc_file(i);
	}

	npc_box_optimize();

	npc_counter_init();
}



/*
 * loads a NPC file
 */
static void load_npc_file(FILE_INDEX file_index)
{
	char description[DESCRIPTION_SIZE];
	CHARACTER *npc = NULL;

        rdb_open(data_path(), RDB_READ);

	while (rdb_next_field()) {
		const char *field_name;

                field_name = rdb_field_name();

		if (strings_equal(field_name, "NAME")) {

			if (npc != NULL) {

				load_npc_pp(npc);
			}

			npc = npc_box_new_npc();

			clear_string(description);

			read_name_field(npc->name,
				CHARACTER_NAME_SIZE
			);

			npc->source_file = file_index;

		} else if (strings_equal(field_name, "CAREER")) {

			read_career_field(npc);

		} else if (strings_equal(field_name, "HOSTILE")) {

			read_boolean_field(&npc->is_hostile);

		} else if (strings_equal(field_name, "GENDER")) {

			read_gender_field(&npc->gender);

		} else if (strings_equal(field_name, "SYMBOL")) {

			read_symbol_field(&npc->symbol);

		} else if (strings_equal(field_name, "COLOR")) {

			npc->color = name_to_color(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (is_short_stat_name(field_name)) {

			read_stat_field(npc, field_name);

		} else if (strings_equal(field_name, "OBJECT")) {
			OBJECT *object;

			object = read_object_field();

			inventory_add(npc, object);

		} else if (strings_equal(field_name, "PERK")) {
			PERK perk;

			perk = name_to_perk(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

			give_perk(npc, perk);

		} else if (strings_equal(field_name, "PSYCHIC_POWER")) {
			PSY_POWER power;

			power = name_to_psy_power(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

			npc->psy_power[power] = true;

		} else if (strings_equal(field_name, "STATUS")) {
			CHARACTER_FLAG flag;

			flag = name_to_character_flag(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

			character_set_flag(npc, flag);

		} else if (strings_equal(field_name, "AI_DEFAULT_STATE")) {

			read_ai_state_field(npc);

		} else if (strings_equal(field_name, "RECRUITABLE")) {

			read_boolean_field(&npc->recruitable);


		} else if (strings_equal(field_name, "DESCRIPTION")) {

			read_description_field(description, npc->name);

		} else {

			die("*** CORE ERROR *** invalid field name: %s (%s)",
				field_name,
				data_path()
			);
		}
        }

	if (npc != NULL) {

		load_npc_pp(npc);
	}

	rdb_close();
}



/*
 * NPC loading post-processor
 */
static void load_npc_pp(CHARACTER *npc)
{

	equip_objects(npc);

	if (npc->perk[PK_STEALTH]) {

		activate_stealth(npc);
	}
}



/*
 * loads the world file
 */
static void load_world_file(void)
{
	WORLD_POINT current_position = {0, 0, 0};

	set_data_path(DIR_WORLD, FILE_WORLD);

	rdb_open(data_path(), RDB_READ);

	while (rdb_next_field()) {
		const char *field_name;

                field_name = rdb_field_name();
		
		if (strings_equal(field_name, "TILE")) {

			read_map_tile();
			
		} else if (strings_equal(field_name, "M")) {

			read_map_field(&current_position);

		} else {

			die("*** CORE ERROR *** invalid field name: %s (%s)",
				field_name,
				data_path()
			);
		}

	}

	rdb_close();

	world_tile_box_optimize();
}



/*
 * loads the dungeon files
 */
static void load_dungeon_files(void) 
{
	const MAP_TILE *tile;
	const MAP_TILE_DATA *tile_data;
	WORLD_POINT p;

	for (p.z = 0, p.y = 0; p.y <= WORLD_MAX_Y; p.y++) {
	for (p.x = 0; p.x <= WORLD_MAX_X; p.x++) {

		tile = world_map_tile(&p);
		tile_data = world_map_tile_data(tile);
		
		if (is_empty_string(tile_data->dungeon_file)) {
		
			continue;
		}
		
		load_dungeon_file(tile_data->dungeon_file, &p);
	}
	}
	
	dungeon_box_optimize();
}



/*
 * loads a dungeon file
 */
static void load_dungeon_file(const char *dungeon_file, 
	const WORLD_POINT *position
)
{
	DUNGEON *dungeon;
	WORLD_COORD dungeon_level;

	set_data_path(DIR_WORLD, dungeon_file);
	
	rdb_open(data_path(), RDB_READ);

	dungeon = dungeon_box_new_dungeon();

	dungeon->location = *position;
	
	dungeon_level = 0;
	
	while (rdb_next_field()) {
		const char *field_name;

                field_name = rdb_field_name();

		if (strings_equal(field_name, "NAME")) {
		
			strcpy(dungeon->name, 
				rdb_data_token(RDB_NEXT_TOKEN)
			);
			
		} else if (strings_equal(field_name, "MAP_SYMBOL")) {
			MAP_TILE_DATA *tile_data;
			
			tile_data = world_map_tile_data(
				world_map_tile(position)
			);
			
			tile_data->symbol = 
				rdb_data_token(RDB_NEXT_TOKEN)[0];

		} else if (strings_equal(field_name, "TYPE")) {

			dungeon->type = name_to_dungeon_type(
				rdb_data_token(RDB_NEXT_TOKEN)
			);
			
		} else if (strings_equal(field_name, "N_LEVELS")) {

			dungeon->n_levels = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);
			
		} else if (strings_equal(field_name, "BASE_LEVEL_Z")) {

			dungeon->base_level_z = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);
			
		} else if (strings_equal(field_name, "N_NPCS_PER_LEVEL")) {

			dungeon->n_npcs_per_level = atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (strings_equal(field_name, "LEVEL")) {

			dungeon_level = dungeon->base_level_z + atoi(
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else if (strings_equal(field_name, "NPC")) {

			spawn_list_add(dungeon, dungeon_level,
				rdb_data_token(RDB_NEXT_TOKEN)
			);

		} else {

			die("*** CORE ERROR *** invalid field name: %s (%s)",
				field_name,
				data_path()
			);
		}
	}

	rdb_close();
}



/*
 * reads symbol field data
 */
static void read_symbol_field(SYMBOL *symbol)
{

	*symbol = rdb_data_token(0)[0];	
}



/*
 * reads dynamic color field data
 */
static void read_dynamic_color_field(DYNAMIC_COLOR *color)
{

	color->primary = name_to_color(rdb_data_token(0));

	if (rdb_optional_data_token(1) == NULL) {
		
		color->secondary = C_NIL;
		
		color->frequency = 0;

		return;
	}

	color->secondary = name_to_color(rdb_data_token(1));

	color->frequency = atoi(rdb_data_token(2));
}



/*
 * reads stat field data
 */
static void read_stat_field(CHARACTER *character, const char *field_name)
{
	STAT i;

	i = name_to_stat(field_name, true);

	character->stat[i].current = character->stat[i].total = 
		atoi(rdb_data_token(0));

	if (rdb_optional_data_token(1) != NULL) {

		character->stat[i].advance = atoi(rdb_data_token(1));
	}
}



/*
 * reads firing mode field data
 */
static void read_firing_mode_field(OBJECT_DATA *object_data)
{
	FIRING_MODE firing_mode;

	firing_mode = name_to_firing_mode(rdb_data_token(RDB_NEXT_TOKEN));

	object_data->firing_mode.has[firing_mode] = true;

	if (firing_mode == FMODE_S) {

		return;
	}

	if (firing_mode == FMODE_SA) {

		object_data->firing_mode.min_sa_shots =
			atoi(rdb_data_token(RDB_NEXT_TOKEN));
		
		object_data->firing_mode.max_sa_shots =
			atoi(rdb_data_token(RDB_NEXT_TOKEN));

	} else if (firing_mode == FMODE_A) {

		object_data->firing_mode.a_shots =
			atoi(rdb_data_token(RDB_NEXT_TOKEN));
	}
}



/*
 * reads dice roll field data
 */
static void read_dice_roll_field(DICE_ROLL *dice_roll)
{

	dice_roll->n_dice = atoi(rdb_data_token(RDB_NEXT_TOKEN));

	dice_roll->n_sides = atoi(rdb_data_token(RDB_NEXT_TOKEN));

	if (rdb_optional_data_token(2) == NULL) {
		
		dice_roll->modifier = 0;

		return;
	}

	dice_roll->modifier = atoi(rdb_data_token(2));
}



/*
 * reads object field data
 */
static OBJECT * read_object_field(void)
{
	OBJECT *object;

	object = object_create(rdb_data_token(0));

	if (rdb_optional_data_token(1) == NULL) {

		return object;
	}

	object->charge = atoi(rdb_data_token(1));

	return object;
}



/*
 * reads career field data
 */
static void read_career_field(CHARACTER *character)
{
	CAREER_INDEX career_index;

	career_index = name_to_career_index(
		rdb_data_token(RDB_NEXT_TOKEN)
	);

	career_set(character, career_index);
}



/*
 * reads description field data
 */
static void read_description_field(char *description, const char *entry_name)
{
	char text_buffer[TEXT_BUFFER_SIZE];

	strcpy(text_buffer, description);
	strcat(text_buffer, rdb_data_token(RDB_NEXT_TOKEN));

	if (strlen(text_buffer) >= DESCRIPTION_SIZE) {

		die("*** CORE ERROR *** description of %s in %s "\
			"is too long (limit: %d characters)",
			entry_name, data_path(), DESCRIPTION_SIZE - 1
		);
	}

	strcpy(description, text_buffer);
}



/*
 * reads key field data
 */
static void read_key_field(KEY_CODE *key)
{

	*key = name_to_key_code(rdb_data_token(RDB_NEXT_TOKEN));
}



/*
 * reads option value field data
 */
static void read_option_value_field(OPTION_VALUE *option_value)
{

	*option_value = (OPTION_VALUE)(
		atoi(rdb_data_token(RDB_NEXT_TOKEN))
	);
}



/*
 * reads boolean field data
 */
static void read_boolean_field(bool *data)
{

	if (strings_equal(rdb_data_token(0), "Yes")) {

		*data = true;

	} else if (strings_equal(rdb_data_token(0), "No")) {

		*data = false;

	} else {

		die("*** CORE ERROR *** invalid data in " \
			"boolean field: %s", rdb_data_token(0)
		);
	}

}



/*
 * reads AI state field data
 */
static void read_ai_state_field(CHARACTER *character)
{

	if (strings_equal(rdb_data_token(0), "Wait")) {

		character->ai.state =
			character->ai.default_state =
			character->ai.original_state =
			AI_STATE_WAIT;

	} else {

		die("*** CORE ERROR *** invalid data in " \
			"AI state field: %s", rdb_data_token(0)
		);
	}

}



/*
 * reads gender field data
 */
static void read_gender_field(GENDER *gender)
{

	switch (rdb_data_token(0)[0]) {

	case 'M':
		*gender = GENDER_MALE;
		break;
	
	case 'F':
		*gender = GENDER_FEMALE;
		break;
	
	case 'N':
		*gender = GENDER_NEUTER;
		break;
	
	default:
		die("*** CORE ERROR *** invalid data in gender field: %c",
			rdb_data_token(0)[0]
		);
	}
}


/*
 * reads name field data
 */
static void read_name_field(char *name, unsigned int n)
{

	if (strlen(rdb_data_token(0)) < n) {
		
		strcpy(name, rdb_data_token(0));
		
		return;
	}

	die("*** CORE ERROR *** name too long: %s " \
		"(limit: %d characters)",
		rdb_data_token(0), n - 1
	);
}



/*
 * reads career exit field data
 */
static void read_career_exit_field(CAREER *career, 
	CAREER_EXIT_INDEX career_exit_index
)
{
	const char *career_name;

	career_name = rdb_data_token(RDB_NEXT_TOKEN);

	strcpy(career->exit[career_exit_index], career_name);
}



/*
 * reads map tile data
 */
static void read_map_tile(void)
{
	MAP_TILE_DATA *tile_data;
	const char *dungeon_file;
	
	tile_data = world_tile_box_new_tile();
	
	tile_data->symbol = rdb_data_token(RDB_NEXT_TOKEN)[0];
	strcpy(tile_data->name, rdb_data_token(RDB_NEXT_TOKEN));
	tile_data->color = name_to_color(rdb_data_token(RDB_NEXT_TOKEN));
	
	dungeon_file = rdb_optional_data_token(RDB_NEXT_TOKEN);
	
	if (dungeon_file != NULL) {
	
		strcpy(tile_data->dungeon_file, dungeon_file);
	}
}



/*
 * reads map field data
 */
static void read_map_field(WORLD_POINT *current_position)
{
	const char *map_line;
	
	map_line = rdb_data_token(RDB_NEXT_TOKEN);
	
	current_position->x = 0;

	while (*map_line != '\0') {
		MAP_TILE *tile;
		
		tile = world_map_tile(current_position);
		
		*tile = symbol_to_world_map_tile(*map_line);

		++map_line;
		++current_position->x;
	}

	++current_position->y;
}



