/*
 * Copyright (C) 2002-2007 The Warp Rogue Team
 * Part of the Warp Rogue Project
 *
 * This software is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License.
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.
 *
 * See the license.txt file for more details.
 */

/*
 * Module Name: Dungeon
 * Description: -
 */

/*
 * Hive city
 */
 
#define Uses_Area
#define Uses_Sector
#define Uses_Object
#define Uses_Terrain
#define Uses_ProgramManager
#define Uses_LoadSave
#define Uses_Util
#include "mheader.h"

#include "dungeon.h"

#include "d_hwrld.h"

#include "d_hcity.h"

#include "cellpnt.h"


#define MIN_SLUDGE_POOLS		2
#define MAX_SLUDGE_POOLS		6
#define MIN_SLUDGE_POOL_EXTENSION	0
#define MAX_SLUDGE_POOL_EXTENSION	5

#define HIVE_CITY_Y_OFFSET		1
#define HIVE_CITY_X_OFFSET		8
#define HIVE_CITY_GATE_HEIGHT		5
#define HIVE_CITY_GATE_SIDE_EX		2


static void		hive_city_generate_level(const DUNGEON *);

static void		hive_city_generate_sector(const AREA_POINT *, 
				CPN_TILE
			);

static void		hive_city_build_gate(void);

static void		add_elevator(void);

static void		add_chemical_sludge(void);
static void		add_chemical_sludge_pool(void);



/*
 * the area section occupied by the hive city
 */
static const AREA_SECTION	HiveCityArea = {
	
	HIVE_CITY_Y_OFFSET, 61,
	HIVE_CITY_X_OFFSET, 118

};



/*
 * generates the hive city
 */
void hive_city_generate(const DUNGEON *dungeon)
{
	DUNGEON_LEVEL level;
	
	for (level = 0; level < dungeon->n_levels; level++) {
		AREA *area;
		

		area = area_new();
		
		area->location = dungeon->location;
		area->location.z = dungeon->base_level_z + level;

		set_active_area(area);

		hive_city_generate_level(dungeon);
		
		spawn_npcs();

		area_write(NULL);

		area_destroy();
	}
}



/*
 * generates a hive city level
 */
static void hive_city_generate_level(const DUNGEON *dungeon)
{
	AREA *area;
	CPN_DUNGEON *dungeon_map;
	int y, x;

	area = active_area();

	strcpy(area->name, dungeon->name);

	dungeon_map = cpn_build_dungeon(5, 5, NULL);

	if (dungeon_map == NULL) {
		
		die(cpn_get_error());
	}

	for (y = 0; y < dungeon_map->height; y++) {
	for (x = 0; x < dungeon_map->width; x++) {
		AREA_POINT p;

		p.y = y + HIVE_CITY_Y_OFFSET;
		p.x = x + HIVE_CITY_X_OFFSET;

		hive_city_generate_sector(&p, dungeon_map->map[y][x]);
	}
	}

	cpn_free_dungeon(dungeon_map);

	if (area->location.z > 0) {

		empty_to_smog();

	} else {

		hive_city_build_gate();

		empty_to_ash_wastes();
	}

	add_elevator();

	add_chemical_sludge();
}



/*
 * generates a hive city sector
 */
static void hive_city_generate_sector(const AREA_POINT *p, CPN_TILE dungeon_tile)
{
	TERRAIN *terrain;
	OBJECT *object;

	object = NULL;

	terrain = terrain_create("Metal floor");
	place_terrain(terrain, p);

	if (dungeon_tile == CPN_FLOOR) return;

	if (dungeon_tile == CPN_WALL) {

		object = object_create("Metal wall");

	} else if (dungeon_tile == CPN_DESTRUCTABLE_OBSTACLE) {

		object = object_create("Barbed wire");

	} else if (dungeon_tile == CPN_DESTRUCTABLE_OBSTACLE_2) {

		object = object_create("Barricade");

	} else {

		die("*** CORE ERROR *** dungeon conversion failed");

	}

	if (object != NULL) {

		place_object(object, p);
	}
}



/*
 * builds the hive city gate
 */
static void hive_city_build_gate(void)
{
	AREA_POINT c;
	int y_mod;
	bool connection_reached;

	c.y = HiveCityArea.bottom;
	c.x = HiveCityArea.left +
		((HiveCityArea.right - HIVE_CITY_X_OFFSET) / 2);

	y_mod = 0;
	connection_reached = false;

	do {
		int x_mod;

		for (x_mod = -(HIVE_CITY_GATE_SIDE_EX);
			x_mod <= HIVE_CITY_GATE_SIDE_EX;
			x_mod++) {
			OBJECT *object;
			AREA_POINT p;

			p.y = c.y + y_mod;
			p.x = c.x + x_mod;

			object = object_at(&p);

			if (object == NULL) {

				connection_reached = true;

			} else {

				object_destroy(object);
				
				sector_at(&p)->object = NULL;
			}
		}

		--y_mod;

	} while (!connection_reached ||
		y_mod > -(HIVE_CITY_GATE_HEIGHT)
	);
}



/*
 * adds the elevator
 */
static void add_elevator(void)
{
	AREA_POINT location;
	OBJECT *elevator;

	sector_set_bounds(&HiveCityArea); 

	if (sector_random(&location, 
		SC_FREE_OBJECT_LOCATION_ANTIBLOCK) == NULL) {

		die("*** CORE ERROR *** could not place elevator");
	}
		
	sector_reset_bounds();

	elevator = object_create("Elevator");

	place_object(elevator, &location);
}



/*
 * adds chemical sludge
 */
static void add_chemical_sludge(void)
{
	int i;

	for (i = random_int(MIN_SLUDGE_POOLS, MAX_SLUDGE_POOLS);
		i > 0;
		--i) {

		add_chemical_sludge_pool();
	}
}



/*
 * adds a chemical sludge pool
 */
static void add_chemical_sludge_pool(void)
{
	TERRAIN *terrain;
	AREA_POINT pool_centre;
	DIRECTION direction[MAX_DIRECTIONS];
	int i, j;

	if (sector_random(&pool_centre, SC_FREE) == NULL) {

		return;
	}

	terrain_destroy(terrain_at(&pool_centre));
	sector_at(&pool_centre)->terrain = NULL;

	terrain = terrain_create("Chemical sludge");
	place_terrain(terrain, &pool_centre);

	randomized_directions(direction);

	for (i = random_int(MIN_SLUDGE_POOL_EXTENSION,
		MAX_SLUDGE_POOL_EXTENSION), j = 0;
		i > 0;
		--i, ++j) {
		AREA_POINT pool_extension;

		pool_extension = pool_centre;
		move_area_point(&pool_extension, direction[j]);

		if (out_of_area_bounds(&pool_extension)) {
			
			continue;
		}

		if (sector_is_class(&pool_extension, SC_FREE)) {

			terrain_destroy(terrain_at(&pool_extension));
			sector_at(&pool_extension)->terrain = NULL;

			terrain = terrain_create("Chemical sludge");
			place_terrain(terrain, &pool_extension);
		}
	}
}



