/*
 * Copyright (C) 2002-2007 The Warp Rogue Team
 * Part of the Warp Rogue Project
 *
 * This software is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License.
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.
 *
 * See the license.txt file for more details.
 */

/*
 * Module Name: Dungeon
 * Description: -
 */

#define Uses_Ui
#define Uses_World
#define Uses_Area
#define Uses_Sector
#define Uses_Terrain
#define Uses_Object
#define Uses_Character
#define Uses_CharacterGeneration
#define Uses_Util
#define Uses_ProgramManager
#define Uses_LoadSave
#define Uses_Npc
#define Uses_Party
#define Uses_DataFile
#include "mheader.h"

#include "dungeon.h"

#include "d_hcity.h"



/*
 * dungeon box buffer size
 */
#define DUNGEON_BOX_BUFFER_SIZE	16



/*
 * number of dungeons
 */
typedef int		N_DUNGEONS;

/*
 * dungeon index
 */
typedef BOX_ITEM_INDEX	DUNGEON_INDEX;



static void		generate_dungeons(void);

static void		spawn_list_new(DUNGEON *);
static LIST *		spawn_list_current(void);

static bool		spawn_npc(NPC_INDEX);
static bool		spawn_unique_npc(NPC_INDEX);
static bool		spawn_unique_npcs(LIST *);

static void *		dungeon_new(void);
static void		dungeon_free(void *);



/*
 * the dungeon box
 */
static BOX *		DungeonBox = NULL;



/*
 * dungeon template (default values)
 */
static const DUNGEON	DungeonTemplate = {
	
	/* NAME */ "", /* TYPE */ DGT_NIL, /* N_LEVELS */ 0,
	/* BASE_LEVEL_Z */ 0, /* N_NPCS_PER_LEVEL */ 0, 
	/* SPAWN LIST */ NULL, /* LOCATION */ {WORLD_COORD_NIL, 
	WORLD_COORD_NIL, WORLD_COORD_NIL}

};




/*
 * dungeon type names
 */
static const char *	DungeonTypeName[MAX_DUNGEON_TYPES] = {

	"Hive city",
	"Cave",
	"Not implemented"

};





/*
 * Dungeon module init
 */
void dungeon_init(void)
{

	DungeonBox = box_new(dungeon_new, dungeon_free,
		DUNGEON_BOX_BUFFER_SIZE
	);
}



/*
 * Dungeon module clean up
 */
void dungeon_clean_up(void)
{

	if (DungeonBox != NULL) {

		box_free(DungeonBox);
	}
}



#if defined(FOR_THE_BLIND)

/*
 * the dungeon generation screen
 */
void dungeon_generation(void)
{
	command_bar_clear();
	render_dungeon_generation_screen(false);
	update_screen();

	generate_dungeons();
	
	command_bar_set(1, CM_OK);
	render_dungeon_generation_screen(true);
	update_screen();
	
	command_bar_get_command();
}

#else 

/*
 * the dungeon generation screen
 */
void dungeon_generation(void)
{
	command_bar_clear();

	render_dungeon_generation_screen(false);
	update_screen();

	generate_dungeons();
}

#endif



/*
 * allocates a new dungeon
 */
DUNGEON * dungeon_box_new_dungeon(void)
{

	return box_new_item(DungeonBox);
}



/*
 * optimizes the dungeon box
 */
void dungeon_box_optimize(void)
{

	box_optimize(DungeonBox);
}



/*
 * returns a pointer to the current dungeon
 */
DUNGEON * dungeon_current(void)
{
	DUNGEON_INDEX i;
	const WORLD_POINT *current_location;
	
	current_location = &active_area()->location;
	
	for (i = 0; i < DungeonBox->current_size; i++) {
		DUNGEON *dungeon;
	
		dungeon = DungeonBox->item[i];

		if (dungeon->location.y == current_location->y &&
			dungeon->location.x == current_location->x) {
			
			return dungeon;
		}
	}

	return NULL;
}



/*
 * (re)spawns the NPCs
 */
void spawn_npcs(void)
{
	LIST *spawn_list;
	LIST_NODE_INDEX i_max;
	N_CHARACTERS n_npcs, n_npcs_present;
	bool only_uniques;

	spawn_list = spawn_list_current();
	if (spawn_list == NULL) return;
	
	only_uniques = spawn_unique_npcs(spawn_list);
	if (only_uniques) return;

	i_max = spawn_list->n_nodes - 1;
	if (i_max < 0) return;
	
	n_npcs_present = area_n_characters() - party_n_members();

	for (n_npcs = dungeon_current()->n_npcs_per_level - n_npcs_present;
		n_npcs > 0; 
		n_npcs--) {
		NPC_INDEX npc_index;
		LIST_NODE_INDEX i;

		i = random_int(0, i_max);

		npc_index = *((NPC_INDEX *)
			list_data_at(spawn_list, i)
		);
			
		if (npc_is_unique(npc_index)) {
				
			++n_npcs;

			continue;
		} 
	
		spawn_npc(npc_index);
	}
}



/*
 * adds a NPC to a spawn list
 */
void spawn_list_add(DUNGEON *dungeon, DUNGEON_LEVEL dungeon_level, 
	const char *npc_name
)
{
	NPC_INDEX *npc_index;

	if (dungeon->spawn_list == NULL) {

		spawn_list_new(dungeon);
	}

	npc_index = checked_malloc(sizeof *npc_index);
	
	if (strings_equal(npc_name, "RANDOM_HUMAN")) {
		
		*npc_index = NPC_RANDOM_HUMAN;

	} else {
		
		*npc_index = name_to_npc_index(npc_name);
	}

	if (dungeon->spawn_list[dungeon_level] == NULL) {

		dungeon->spawn_list[dungeon_level] = list_new();
	}

	list_add(dungeon->spawn_list[dungeon_level], npc_index);
}



/*
 * name -> dungeon type
 */
DUNGEON_TYPE name_to_dungeon_type(const char *name)
{
	DUNGEON_TYPE i;

	for (i = 0; i < MAX_DUNGEON_TYPES; i++) {

		if (strings_equal(name, DungeonTypeName[i])) {

			return i;
		}
	}

	die("*** CORE ERRROR *** invalid dungeon type: %s", name);

	/* NEVER REACHED */
	return DGT_NIL;
}



/*
 * generates the dungeons
 */
static void generate_dungeons(void)
{
	DUNGEON_INDEX i;

	for (i = 0; i < DungeonBox->current_size; i++) {
		const DUNGEON *dungeon;
		
		dungeon = DungeonBox->item[i];

		switch (dungeon->type) {
			
		case DGT_HIVE_CITY:
			hive_city_generate(dungeon);
			break;
		
		case DGT_NOT_IMPLEMENTED:
			/* DO NOTHING */
			break;
		
		default:
			die("*** CORE ERROR *** invalid dungeon type"); 
		}
	}
}



/*
 * allocates memory for the spawn list of a dungeon 
 */
static void spawn_list_new(DUNGEON *dungeon)
{
	DUNGEON_LEVEL i;

	dungeon->spawn_list = checked_malloc(
		dungeon->n_levels * sizeof *dungeon->spawn_list
	);

	for (i = 0; i < dungeon->n_levels; i++) {

		dungeon->spawn_list[i] = NULL;
	}
}



/* 
 * returns the current spawn list, or NULL if there is no spawn list
 */
static LIST * spawn_list_current(void)
{
	DUNGEON *dungeon;
	LIST *spawn_list;
	
	dungeon = dungeon_current();

	if (dungeon == NULL || dungeon->spawn_list == NULL) {
	
		return NULL;
	}
			
	spawn_list = dungeon->spawn_list[
		active_area()->location.z - 
		dungeon->base_level_z
	];
			
	return spawn_list;
}



/*
 * spawns an NPC
 */
static bool spawn_npc(NPC_INDEX npc_index)
{
	CHARACTER *npc;
	AREA_POINT spawn_point;
	
	if (sector_random(&spawn_point, SC_MOVE_TARGET_SAFE) == NULL) {
		
		return false;
	}

	if (npc_index == NPC_RANDOM_HUMAN) {
			
		npc = create_random_human();
		
	} else {
			
		npc = npc_create_i(npc_index);
	}

	place_character(npc, &spawn_point);
	
	return true;
}



/*
 * spawns a unique NPC
 */
static bool spawn_unique_npc(NPC_INDEX npc_index)
{
	NPC_COUNTER *counter;
			
	counter = npc_counter(npc_index);
		
	if (counter->n_created != 0) {

		return false;
	}

	spawn_npc(npc_index);

	counter->n_created += 1;

	return true;
}



/*
 * spawns the unique NPCs
 *
 * returns true if there are only uniques on the passed spawn list
 */
static bool spawn_unique_npcs(LIST *spawn_list)
{
	LIST_NODE *node;
	bool only_uniques;

	only_uniques = true;

	for (node = spawn_list->head; node != NULL; node = node->next) {
		NPC_INDEX npc_index;

		npc_index = *((NPC_INDEX *)node->data);
			
		if (!npc_is_unique(npc_index)) {

			only_uniques = false;
			
			continue;
		}

		spawn_unique_npc(npc_index);
	}

	return only_uniques;
}



/*
 * allocs a dungeon
 */
static void * dungeon_new(void)
{
	DUNGEON *dungeon;
	
	dungeon = checked_malloc(sizeof *dungeon);
	
	*dungeon = DungeonTemplate;

	return dungeon;
}



/*
 * frees a dungeon
 */
static void dungeon_free(void *data)
{
	DUNGEON *dungeon;
	DUNGEON_LEVEL i;
	
	dungeon = (DUNGEON *)data;

	if (dungeon->spawn_list == NULL) {
		
		free(dungeon);
		
		return;
	}

	for (i = 0; i < dungeon->n_levels; i++) {

		if (dungeon->spawn_list[i] == NULL) {
		
			continue;
		}
		
		list_destroy_with(dungeon->spawn_list[i], free);
	}

	free(dungeon->spawn_list);

	free(dungeon);
}

