/*
 * Copyright (C) 2002-2007 The Warp Rogue Team
 * Part of the Warp Rogue Project
 *
 * This software is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License.
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.
 *
 * See the license.txt file for more details.
 */

/*
 * Module Name: DynamicMessage
 * Description: This module is responsible for dynamic messages
 *              i.e. "You hit the ork.", "The ork hits you.", etc.
 */

#include <ctype.h>

#define Uses_Ui
#define Uses_Grammar
#define Uses_Util
#define Uses_Object
#define Uses_Character
#define Uses_Perception
#include "mheader.h"

#include "dyn_msg.h"


#define MAX_MESSAGE_TYPES		34


/*
 * message data structure
 */
typedef struct {

	/* the message function */
	void		(* const send)(const CHARACTER *, const void *, 
				MESSAGE_OBJECT_TYPE
			);

	/* true if the message function accepts unnoticed subjects */
	const bool	unnoticed_variant;

} MESSAGE_DATA;



static void	message_hit(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_critical_hit(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_miss(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_parry(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_disarm(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_disarm_fails(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);
		
static void	message_doctor(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_doctor_fails(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_pick_up(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_reload_weapon(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_recharge_weapon(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_switch_weapons(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_fights_unarmed(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_weapon_jams(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_unjam_weapon(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_weapon_malfunctions(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_weapon_overheats(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_weapon_explodes(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_resist(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_recover(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_effect_activation(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_effect_termination(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_concentrate(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_evoke_psy_power(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_evoke_psy_power_fails(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_psychic_overload(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_psy_power_termination(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_use_drug(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_drug_termination(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_fade_into_shadows(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);
		
static void	message_notice(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_fall_down(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_death(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);

static void	message_destruction(const CHARACTER *, const void *,
			MESSAGE_OBJECT_TYPE
		);



/*
 * table of message functions
 */
static  MESSAGE_DATA Message[MAX_MESSAGE_TYPES] = {

	{message_hit,			true},
	{message_critical_hit,		true},
	{message_miss,			true},
	{message_parry,			false},

	{message_disarm,		true},
	{message_disarm_fails,		true},
	
	{message_doctor,		true},
	{message_doctor_fails,		true},

	{message_pick_up,		false},

	{message_reload_weapon,		false},
	{message_recharge_weapon,	false},
	{message_switch_weapons,	false},
	{message_fights_unarmed,	false},
	{message_weapon_jams,		false},
	{message_unjam_weapon,		false},
	
	{message_weapon_malfunctions,	false},

	{message_weapon_overheats,	false},
	{message_weapon_explodes,	false},

	{message_resist,		false},
	{message_recover,		false},

	{message_effect_activation,	false},
	{message_effect_termination,	false},

	{message_concentrate,		false},
	{message_evoke_psy_power,	false},
	{message_evoke_psy_power_fails,	false},
	{message_psychic_overload,	false},
	{message_psy_power_termination,	false},

	{message_use_drug,		false},
	{message_drug_termination,	false},

	{message_fade_into_shadows,	false},
	{message_notice,		false},

	{message_fall_down,		false},

	{message_death,			false},
	{message_destruction,		false}

};



/*
 * true if the subject of the message has not been noticed by
 * the player's party
 */
static bool SubjectUnnoticed;




/*
 * outputs a dynamic message
 */
void dynamic_message(MESSAGE_TYPE message_type, const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{

	if (subject != NULL) {

		if (party_can_see(subject)) {
			
			SubjectUnnoticed = false;
		
		} else {
		
			SubjectUnnoticed = true;
		}

		if (SubjectUnnoticed && 
			!(Message[message_type].unnoticed_variant)) {
		
			return;
		}
	}

	(Message[message_type].send)(subject, object, object_type);
}



/*
 * hit message
 */
static void message_hit(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{

	if (SubjectUnnoticed) {
		
		message(SID_GAME, "Something hits %s.",
			object_string(object, object_type, ARTICLE_DEFINITE)
		);
	
		return;
	}

	message(SID_GAME, "%s %s %s.",
		subject_string(subject),
		conjugated_verb("hit", grammatical_person(subject)),
		object_string(object, object_type, ARTICLE_DEFINITE)
	);
}



/*
 * critical hit message
 */
static void message_critical_hit(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{

	if (SubjectUnnoticed) {
		
		message(SID_GAME, "Something *critically* hits %s.",
			object_string(object, object_type, ARTICLE_DEFINITE)
		);
	
		return;
	}
	
	message(SID_GAME, "%s *critically* %s %s.",
		subject_string(subject),
		conjugated_verb("hit", grammatical_person(subject)),
		object_string(object, object_type, ARTICLE_DEFINITE)
	);
}



/*
 * miss message
 */
static void message_miss(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{

	if (SubjectUnnoticed) {
		
		message(SID_GAME, "Something misses %s.",
			object_string(object, object_type, ARTICLE_DEFINITE)
		);
	
		return;
	}

	message(SID_GAME, "%s %s %s.",
		subject_string(subject),
		conjugated_verb("miss", grammatical_person(subject)),
		object_string(object, object_type, ARTICLE_DEFINITE)
	);
}



/*
 * parry message
 */
static void message_parry(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{

	message(SID_GAME, "%s %s %s attack.",
		subject_string(subject),
		conjugated_verb("parry", grammatical_person(subject)),
		possessive_form(object_string(object, object_type,
			ARTICLE_DEFINITE)
		)
	);
}



/*
 * disarm message
 */
static void message_disarm(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{

	if (SubjectUnnoticed) {

		message(SID_GAME, "Something disarms %s.",
			object_string(object, object_type, ARTICLE_DEFINITE)
		);
	
		return;
	}

	message(SID_GAME, "%s %s %s.",
		subject_string(subject),
		conjugated_verb("disarm", grammatical_person(subject)),
		object_string(object, object_type, ARTICLE_DEFINITE)
	);
}



/*
 * disarm fails message
 */
static void message_disarm_fails(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	char c_try[STRING_BUFFER_SIZE];
	
	if (SubjectUnnoticed) {
		
		message(SID_GAME, "Something tries to disarm %s but fails.",
			object_string(object, object_type, ARTICLE_DEFINITE)
		);
	
		return;
	}

	strcpy(c_try, conjugated_verb("try", grammatical_person(subject)));

	message(SID_GAME, "%s %s to disarm %s but %s.",
		subject_string(subject),
		c_try,
		object_string(object, object_type, ARTICLE_DEFINITE),
		conjugated_verb("fail", grammatical_person(subject))
	);
}



/*
 * doctor message
 */
static void message_doctor(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	/* hack - fix later */
	if (subject == object) {

		message(SID_GAME, "You treat yourself.");
		
		return;
	}

	if (SubjectUnnoticed) {

		message(SID_GAME, "Something treats %s.",
			object_string(object, object_type, ARTICLE_DEFINITE)
		);
	
		return;
	}

	message(SID_GAME, "%s %s %s.",
		subject_string(subject),
		conjugated_verb("treat", grammatical_person(subject)),
		object_string(object, object_type, ARTICLE_DEFINITE)
	);
}



/*
 * doctor fails message
 */
static void message_doctor_fails(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	char c_try[STRING_BUFFER_SIZE];
	
	/* hack - fix later */
	if (subject == object) {

		message(SID_GAME, "You attempt to treat yourself but fail.");
		
		return;
	}
	
	if (SubjectUnnoticed) {
		
		message(SID_GAME, "Something attempts to treat %s but fails.",
			object_string(object, object_type, ARTICLE_DEFINITE)
		);
	
		return;
	}

	strcpy(c_try, conjugated_verb("attempt", grammatical_person(subject)));

	message(SID_GAME, "%s %s to treat %s but %s.",
		subject_string(subject),
		c_try,
		object_string(object, object_type, ARTICLE_DEFINITE),
		conjugated_verb("fail", grammatical_person(subject))
	);
}




/*
 * pick up message
 */
static void message_pick_up(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{

	message(SID_GAME, "%s %s up %s.",
		subject_string(subject),
		conjugated_verb("pick", grammatical_person(subject)),
		object_string(object, object_type, ARTICLE_INDEFINITE)
	);
}



/*
 * reload weapon message
 */
static void message_reload_weapon(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object);
	NOT_USED(object_type);

	message(SID_GAME, "%s %s.",
		subject_string(subject),
	        conjugated_verb("reload", grammatical_person(subject))
	);
}



/*
 * recharge weapon message
 */
static void message_recharge_weapon(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object);
	NOT_USED(object_type);

	if (subject == NULL) return;

	message(SID_GAME, "%s weapon has recharged.",
		possessive_form(subject_string(subject))
	);
}



/*
 * switch weapons message
 */
static void message_switch_weapons(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{

	if (object == NULL) {

		message_fights_unarmed(subject, object, object_type);

		return;
	}

	message(SID_GAME, "%s %s %s.",
		subject_string(subject),
		conjugated_verb("draw", grammatical_person(subject)),
		object_string(object, object_type, ARTICLE_INDEFINITE)
	);
}



/*
 * fights unarmed message
 */
static void message_fights_unarmed(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object);
	NOT_USED(object_type);

	message(SID_GAME, "%s now %s unarmed.",
		subject_string(subject),
		conjugated_verb("fight", grammatical_person(subject))
	);
}



/*
 * weapon jams message
 */
static void message_weapon_jams(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object);
	NOT_USED(object_type);

	message(SID_GAME, "%s weapon jams.",
		possessive_form(subject_string(subject))
	);
}



/*
 * unjam weapon message
 */
static void message_unjam_weapon(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object);
	NOT_USED(object_type);

	message(SID_GAME, "%s %s %s weapon.",
		subject_string(subject),
		conjugated_verb("unjam", grammatical_person(subject)),
		possessive_pronoun(subject)
	);
}



/*
 * weapon malfunction message
 */
static void message_weapon_malfunctions(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object);
	NOT_USED(object_type);

	message(SID_GAME, "%s weapon malfunctions.",
		possessive_form(subject_string(subject))
	);
}



/*
 * weapon overheats message
 */
static void message_weapon_overheats(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object);
	NOT_USED(object_type);

	message(SID_GAME, "%s weapon overheats.",
		possessive_form(subject_string(subject))
	);

	message(SID_GAME, "%s %s the weapon.",
		subject_string(subject),
		conjugated_verb("drop", grammatical_person(subject))
	);
}



/*
 * weapon explodes message
 */
static void message_weapon_explodes(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object);
	NOT_USED(object_type);

	message(SID_GAME, "%s weapon explodes.",
		possessive_form(subject_string(subject))
	);
}



/*
 * resist message
 */
static void message_resist(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object);
	NOT_USED(object_type);

	message(SID_GAME, "%s %s.",
		subject_string(subject),
		conjugated_verb("resist", grammatical_person(subject))
	);
}



/*
 * recover message
 */
static void message_recover(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object);
	NOT_USED(object_type);

	message(SID_GAME, "%s %s.",
		subject_string(subject),
	        conjugated_verb("recover", grammatical_person(subject))
	);
}



/*
 * effect activation message
 */
static void message_effect_activation(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object_type);

	message(SID_GAME, "%s %s %s.",
		subject_string(subject),
		conjugated_verb("be", grammatical_person(subject)),
		lowercase_object_string(object)
	);
}



/*
 * effect termination message
 */
static void message_effect_termination(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object_type);

	message(SID_GAME, "%s %s no longer %s.",
		subject_string(subject),
		conjugated_verb("be", grammatical_person(subject)),
		lowercase_object_string(object)
	);
}



/*
 * concentrate message
 */
static void message_concentrate(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object);
	NOT_USED(object_type);

	message(SID_GAME, "%s %s.",
		subject_string(subject),
		conjugated_verb("concentrate", grammatical_person(subject))
	);
}



/*
 * psychic power evocation message
 */
static void message_evoke_psy_power(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object_type);

	message(SID_GAME, "%s %s %s.",
		subject_string(subject),
		conjugated_verb("evoke", grammatical_person(subject)),
		lowercase_object_string(object)
	);
}



/*
 * psychic power evocation fails message
 */
static void message_evoke_psy_power_fails(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	char c_try[STRING_BUFFER_SIZE];

	NOT_USED(object_type);

	strcpy(c_try, conjugated_verb("try", grammatical_person(subject)));

	message(SID_GAME, "%s %s to evoke %s but %s.",
		subject_string(subject),
		c_try,
		lowercase_object_string(object),
		conjugated_verb("fail", grammatical_person(subject))
	);
}



/*
 * psychic overload message
 */
static void message_psychic_overload(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object);
	NOT_USED(object_type);

	message(SID_GAME, "%s %s a psychic overload.",
		subject_string(subject),
	        conjugated_verb("suffer", grammatical_person(subject))
	);
}



/*
 * psychic power termination message
 */
static void message_psy_power_termination(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object_type);

	message(SID_GAME, "%s %s no longer under the influence of %s.",
		subject_string(subject),
		conjugated_verb("be", grammatical_person(subject)),
		lowercase_object_string(object)
	);
}



/*
 * use drug message
 */
static void message_use_drug(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object_type);

	message(SID_GAME, "%s %s %s.",
		subject_string(subject),
		conjugated_verb("take", grammatical_person(subject)),
		lowercase_object_string(object)
	);
}



/*
 * drug termination message
 */
static void message_drug_termination(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object_type);

	message(SID_GAME, "%s %s no longer under the influence of %s.",
		subject_string(subject),
		conjugated_verb("be", grammatical_person(subject)),
		lowercase_object_string(object)
	);
}



/*
 * fade into shadows message
 */
static void message_fade_into_shadows(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object);
	NOT_USED(object_type);

	message(SID_GAME, "%s %s into the shadows.",
		subject_string(subject),
		conjugated_verb("fade", grammatical_person(subject))
	);
}



/*
 * notice message
 */
static void message_notice(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{

	message(SID_GAME, "%s %s %s.",
		subject_string(subject),
		conjugated_verb("notice", grammatical_person(subject)),
		object_string(object, object_type, ARTICLE_INDEFINITE)
	);
}



/*
 * fall down message
 */
static void message_fall_down(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object);
	NOT_USED(object_type);

	message(SID_GAME, "%s %s down.",
		subject_string(subject),
		conjugated_verb("fall", grammatical_person(subject))
	);
}



/*
 * death message
 */
static void message_death(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{
	NOT_USED(object);
	NOT_USED(object_type);

	if (character_has_flag(subject, CF_MACHINE)) {

		message(SID_GAME, "%s %s destroyed.",
			subject_string(subject),
			conjugated_verb("be",
				grammatical_person(subject)
			)
		);

	} else {

		message(SID_GAME, "%s %s.",
			subject_string(subject),
			conjugated_verb("die",
				grammatical_person(subject)
			)
		);
	}
}



/*
 * destruction message
 */
static void message_destruction(const CHARACTER *subject,
	const void *object, MESSAGE_OBJECT_TYPE object_type
)
{

	if (subject == NULL) {

		message(SID_GAME, "%s is destroyed.",
			subject_string_o(object)
		);

		return;
	}

	message(SID_GAME, "%s %s is destroyed.",
		possessive_form(subject_string(subject)),
		object_string(object, object_type, ARTICLE_NONE)
	);

}

