/*
 * Copyright (C) 2002-2007 The Warp Rogue Team
 * Part of the Warp Rogue Project
 *
 * This software is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License.
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.
 *
 * See the license.txt file for more details.
 */

/*
 * Module Name: Economy
 * Description: -
 */
 
#define Uses_Ui
#define Uses_Object
#define Uses_Inventory
#define Uses_Perks
#define Uses_Stats
#define Uses_Util
#define Uses_ProgramManager
#include "mheader.h"

#include "economy.h"


/*
 * price modes (in percent)
 */
#define PRICE_NIL			-1
#define PRICE_TRADER			100
#define PRICE_ROGUE_TRADER		300
#define PRICE_SELL			50
#define PRICE_MANUFACTURE		PRICE_SELL



typedef int		PRICE_MODE;



static void		buy_object(CHARACTER *, OBJECT *);
static void		manufacture_object(CHARACTER *, OBJECT *);

static MENU *		object_menu_create(const CHARACTER *, LIST *);
static bool		can_acquire(const CHARACTER *, const OBJECT *, MONEY);
static MONEY		object_price(const CHARACTER *, const OBJECT *);




/*
 * the current price mode
 */
static PRICE_MODE	PriceMode = PRICE_NIL;




/*
 * the buy screen
 */
void buy_screen(CHARACTER *character, PERK used_perk)
{
	MENU *menu;
	LIST *available_objects;

	if (used_perk == PK_TRADING) {

		available_objects = object_list_trader();
		
		PriceMode = PRICE_TRADER;

	} else if (used_perk == PK_ROGUE_TRADING) {

		available_objects = object_list_rogue_trader();

		PriceMode = PRICE_ROGUE_TRADER;
		
	} else {

		return;
	}

	menu = object_menu_create(character, available_objects);

	do  {
		OBJECT *object;

		command_bar_set(5, CM_UP, CM_DOWN, CM_BUY, 
			CM_INFO, CM_EXIT
		);

		render_buy_screen(character, used_perk);

		buy_screen_menu(menu);

		object = (OBJECT *)list_data_at(available_objects,
			menu->highlighted
		);

		if (menu->entered_command == CM_BUY) {

			buy_object(character, object);

		} else if (menu->entered_command == CM_INFO) {

			object_screen(object);
		}

	} while (menu->entered_command != CM_EXIT);

	menu_destroy_with(menu, free);

	list_destroy_with(available_objects, object_destroy);	
}



/*
 * the sell screen
 */
void sell_screen(CHARACTER *character)
{

	inventory_screen(character, true);
}



/*
 * sells an object
 */
void sell_object(CHARACTER *character, OBJECT *object)
{

	inventory_money_increase(character, sell_price(character, object));
}



/*
 * returns the sell price
 */
MONEY sell_price(const CHARACTER *character, const OBJECT *object)
{

	PriceMode = PRICE_SELL;
	
	return object_price(character, object);
}



/*
 * the manufacture screen
 */
void manufacture_screen(CHARACTER *character, PERK used_perk)
{
	MENU *menu;
	LIST *available_objects;

	PriceMode = PRICE_MANUFACTURE;

	available_objects = object_list_drugs(NULL);

	menu = object_menu_create(character, available_objects);

	do  {
		OBJECT *object;

		command_bar_set(5, CM_UP, CM_DOWN, CM_MANUFACTURE, 
			CM_INFO, CM_EXIT
		);

		render_manufacture_screen(character, used_perk);

		manufacture_screen_menu(menu);

		object = (OBJECT *)list_data_at(available_objects,
			menu->highlighted
		);

		if (menu->entered_command == CM_MANUFACTURE) {

			manufacture_object(character, object);

		} else if (menu->entered_command == CM_INFO) {

			object_screen(object);
		}

	} while (menu->entered_command != CM_EXIT);

	menu_destroy_with(menu, free);

	list_destroy_with(available_objects, object_destroy);
}



/*
 * buy object
 */
static void buy_object(CHARACTER *character, OBJECT *object) 
{
	MONEY price;

	price = object_price(character, object);

	if (!can_acquire(character, object, price)) return;

	message(SID_DIALOG, "Bought");

	inventory_money_reduce(character, price);

	inventory_add(character, object_clone(object));
}



/*
 * manufacture object
 */
static void manufacture_object(CHARACTER *character, OBJECT *object)
{
	MONEY price;

	price = object_price(character, object);

	if (!can_acquire(character, object, price)) return;

	message(SID_DIALOG, "Built");

	inventory_money_reduce(character, price);

	inventory_add(character, object_clone(object));
}



/*
 * creates an object menu
 */
static MENU * object_menu_create(const CHARACTER *character,
	LIST *available_objects
){
	LIST *menu_items;
	MENU *menu;
	LIST_NODE *node;

	menu_items = list_new();

	for (node = available_objects->head;
		node != NULL;
		node = node->next) {
		OBJECT *object;
		char *entry;

		object = (OBJECT *)node->data;

		entry = checked_malloc(STRING_BUFFER_SIZE);

		sprintf(entry, "%s (%dc) ",
			object_static_data(object)->name,
			object_price(character, object)
		);

		list_add(menu_items, entry);
	}

	menu = menu_create(menu_items);

	return menu;
}



/*
 * returns true if the passed character can acquire the passed objects
 */
static bool can_acquire(const CHARACTER *character, const OBJECT *object, 
	MONEY price
) {
	
	if (price > inventory_money(character)) {

		message(SID_DIALOG, "Not enough credits");
		
		return false;
	}

	if (inventory_is_full(character)) {

		message(SID_DIALOG, "Inventory is full");
		
		return false;
	}

	if (object_static_data(object)->weight +
		inventory_weight(character) > ENCUMBRANCE_MAX(character)) {

		message(SID_DIALOG, "You cannot carry any more");
			
		return false;
	}

	return true;
}



/*
 * returns the price of an object (dependent on price mode)
 */
static MONEY object_price(const CHARACTER *character, const OBJECT *object)
{
	MONEY value, price;
	
	value = object_static_data(object)->value;
	
	if (value == VALUE_NIL) return value;
	
	price = percent(PriceMode, value);

	if (character->perk[PK_PERSUADE] && (PriceMode == PRICE_TRADER || 
		PriceMode == PRICE_ROGUE_TRADER)) {

		price = percent(PERSUADE_PRICE, price);
	}

	return price;
}


