/*
 * Copyright (C) 2002-2007 The Warp Rogue Team
 * Part of the Warp Rogue Project
 *
 * This software is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License.
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.
 *
 * See the license.txt file for more details.
 */

/*
 * Module Name: Object
 * Description: -
 */

#define Uses_Ui
#define Uses_Util
#define Uses_ProgramManager
#define Uses_DataFile
#define Uses_Area
#define Uses_Event
#include "mheader.h"

#include "object.h"



#define OBJECT_BOX_BUFFER_SIZE	64




static OBJECT_INDEX	name_to_object_index(const char *);

static void *		object_data_new(void);
static void		object_data_free(void *);




/*
 * object type names
 */
static const char *	ObjectTypeName[MAX_OBJECT_TYPES] = {

	"Close combat weapon",
	"Ranged combat weapon",

	"Armour",

	"Drug",

	"Environment",

	"Money",

	"Jump pack",

	"Tome"

};



/*
 * object subtype names
 */
static const char *	ObjectSubtypeName[MAX_OBJECT_SUBTYPES] = {

	"Knife",

	"Pistol",
	"Basic",

	"Light",
	"Medium",
	"Heavy",

	"Ancient tome",
	"Warp tome",

	"Elevator"

};



/*
 * object attribute names
 */
static const char *	ObjectAttributeName[MAX_OBJECT_ATTRIBUTES] = {

	"Unique",
	"Daemonic",
	"Impassable",
	"Dangerous",
	"Blocks line of sight",
	"Shock",
	"Pain",
	"Power",
	"Force",
	"Projectile",
	"Laser",
	"Plasma",
	"Plague",
	"Poison",
	"Silent",
	"Strength boost",
	"Strength boost II",
	"Environmental protection",
	"Force rune",
	"Force channel",
	"Ignore armour",
	"Hinders movement",
	"Partial reload",
	"Automatic recharge",
	"Psychic bolt",
	"Alien",
	"Reliable",
	"Noisy",
	"Inherent",
	"No parry"

};



/*
 * the object box
 */
static BOX *		ObjectBox = NULL;


/*
 * object data template (default values)
 */
static const OBJECT_DATA ObjectDataTemplate = {

	/* NAME */ "", /* ATTRIBUTES */ {false,false,false,false,false,false,
	false,false,false,false,false,false,false,false,false,false,false,
	false,false,false,false,false,false,false,false,false,false,false,
	false,false}, /* SYMBOL */ '!', /* COLOR */ {C_DEFAULT, C_NIL, 0},
	/* CONDITION MAX */ CONDITION_INDESTRUCTABLE,
	/* CHARGE MAX */ CHARGE_NIL, /* TYPE */ OTYPE_NIL,
	/* SUBTYPE */ OSTYPE_NIL, /* WEIGHT */ WEIGHT_NIL,
	/* VALUE */ VALUE_NIL, /* BUYABLE */ false, /* DAMAGE */ {0, 0, 0},
	/* FIRING MODE DATA */ {{false, false, false}, 0, 0, 0, 0},
	/* RELOAD VALUE */ 0, /* REACH */ 0, /* PARRY PENALTY */ 0,
	/* ARMOUR VALUE */ 0, /* SOURCE FILE */ FILE_INDEX_NIL

};



/*
 * Object module init
 */
void object_init(void)
{

	ObjectBox = box_new(object_data_new, object_data_free,
		OBJECT_BOX_BUFFER_SIZE
	);
}



/*
 * Object module clean up
 */
void object_clean_up(void)
{

	if (ObjectBox != NULL) {

		box_free(ObjectBox);
	}
}



/*
 * adds an object to the object box
 */
OBJECT_DATA * object_box_new_object(void)
{

	return box_new_item(ObjectBox);
}



/*
 * optimizes the object box
 */
void object_box_optimize(void)
{

	box_optimize(ObjectBox);
}



/*
 * returns the static data of an object
 */
const OBJECT_DATA * object_static_data(const OBJECT *object)
{
	const OBJECT_DATA *object_data;

	object_data = ObjectBox->item[object->index];

	return object_data;
}



/*
 * returns true if the passed object has the passed attribute
 */
bool object_has_attribute(const OBJECT *object, OBJECT_ATTRIBUTE attribute)
{
	const OBJECT_DATA *object_data;
	
	object_data = ObjectBox->item[object->index];

	return object_data->attribute[attribute];
}



/*
 * creates an object (name based)
 */
OBJECT * object_create(const char *object_name)
{
	OBJECT_INDEX object_index;
	OBJECT *object;

	object_index = name_to_object_index(object_name);

	object = object_create_i(object_index);

	return object;
}



/*
 * creates an object (index based)
 */
OBJECT * object_create_i(OBJECT_INDEX object_index)
{
	OBJECT *object;
	const OBJECT_DATA *object_data;

	object_data = ObjectBox->item[object_index];

	object = checked_malloc(sizeof *object);

	object->index = object_index;
	object->color = dynamic_color(&object_data->color);
	object->gore_level = GORE_LEVEL_ZERO;
	object->condition = object_data->condition_max;
	object->charge = object_data->charge_max;
	object->functional = true;
	object->list_node = NULL;
	object->list_index = LIST_NODE_INDEX_NIL;

	return object;
}



/*
 * destroys an object
 */
void object_destroy(void *p)
{
	OBJECT *object;

	object = (OBJECT *)p;

	if (object->list_node != NULL) {
		LIST *object_events;

		object_events = list_new();

		remove_object(object, object_events);

		list_destroy_with(object_events, event_destroy);
	}

	free(object);
}




/*
 * produces an identical copy of an object
 */
OBJECT * object_clone(const OBJECT *object)
{
	OBJECT *clone;

	clone = checked_malloc(sizeof *clone);

	*clone = *object;

	return clone;
}



/*
 * the object screen
 */
void object_screen(const OBJECT *object)
{
	command_bar_set(1, CM_EXIT);

	render_object_screen(object);

	update_screen();

	command_bar_get_command();
}



/*
 * returns the description of an object
 */
char * object_description(char *description, const OBJECT *object)
{

	return data_file_object_description(description, object);
}



/*
 * returns the name of an object type
 */
const char * object_type_name(OBJECT_TYPE object_type)
{

	return ObjectTypeName[object_type];
}



/*
 * returns the name of an object subtype
 */
const char * object_subtype_name(OBJECT_SUBTYPE object_subtype)
{

	return ObjectSubtypeName[object_subtype];
}



/*
 * returns the name of an object attribute
 */
const char * object_attribute_name(OBJECT_ATTRIBUTE object_attribute)
{

	return ObjectAttributeName[object_attribute];
}



/*
 * name -> object type
 */
OBJECT_TYPE name_to_object_type(const char *name)
{
	OBJECT_TYPE i;

	for (i = 0; i < MAX_OBJECT_TYPES; i++) {

		if (strings_equal(name, ObjectTypeName[i])) {

			return i;
		}
	}

	die("*** CORE ERRROR *** invalid object type: %s", name);

	/* NEVER REACHED */
	return OTYPE_NIL;
}



/*
 * name -> object subtype
 */
OBJECT_SUBTYPE name_to_object_subtype(const char *name)
{
	OBJECT_SUBTYPE i;

	for (i = 0; i < MAX_OBJECT_SUBTYPES; i++) {

		if (strings_equal(name, ObjectSubtypeName[i])) {

			return i;
		}
	}

	die("*** CORE ERROR *** invalid object subtype: %s", name);

	/* NEVER REACHED */
	return OSTYPE_NIL;
}



/*
 * name -> object attribute
 */
OBJECT_ATTRIBUTE name_to_object_attribute(const char *name)
{
	OBJECT_ATTRIBUTE i;

	for (i = 0; i < MAX_OBJECT_ATTRIBUTES; i++) {

		if (strings_equal(name, ObjectAttributeName[i])) {

			return i;
		}
	}

	die("*** CORE ERROR *** invalid object attribute %s", name);

	/* NEVER REACHED */
	return OA_NIL;
}



/*
 * returns a list of all objects which can be bought from a trader
 */
LIST * object_list_trader(void)
{
	const OBJECT_DATA *object_data;
	LIST *list;
	OBJECT_INDEX i;

	list = list_new();

	for (i = 0; i < ObjectBox->current_size; i++) {

		object_data = ObjectBox->item[i];

		if (object_data->value == VALUE_NIL ||
			object_data->attribute[OA_INHERENT] ||
			!object_data->buyable) {

			continue;
		}

		list_add(list, object_create_i(i));
	}

	return list;
}



/*
 * returns a list of all objects which can be bought from a rogue trader
 */
LIST * object_list_rogue_trader(void)
{
	const OBJECT_DATA *object_data;
	LIST *list;
	OBJECT_INDEX i;

	list = list_new();

	for (i = 0; i < ObjectBox->current_size; i++) {

		object_data = ObjectBox->item[i];

		if (object_data->buyable ||
			object_data->value == VALUE_NIL ||
			object_data->attribute[OA_INHERENT] ||
			object_data->attribute[OA_UNIQUE]) {

			continue;
		}

		list_add(list, object_create_i(i));
	}

	return list;
}



/*
 * if character equals NULL a list of all drug objects is returned,
 * otherwise a list of all drug objects present in the passed
 * character's inventory is returned
 */
LIST * object_list_drugs(CHARACTER *character)
{
	LIST *list;
	LIST_NODE *node;

	list = list_new();

	if (character == NULL) {
		int i;

		for (i = 0; i < ObjectBox->current_size; i++) {
			const OBJECT_DATA *object_data;

			object_data = ObjectBox->item[i];

			if (object_data->type == OTYPE_DRUG) {

				list_add(list, object_create_i(i));
			}
		}

		return list;
	}

	for (node = character->inventory->head;
		node != NULL;
		node = node->next) {
		OBJECT *object;

		object = (OBJECT *)node->data;

		if (object_static_data(object)->type ==
			OTYPE_DRUG){

			list_add(list, object);
		}
	}

	return list;
}



/*
 * name -> object index
 */
static OBJECT_INDEX name_to_object_index(const char *name)
{
	OBJECT_INDEX i;

	for (i = 0; i < ObjectBox->current_size; i++) {
		const OBJECT_DATA *object_data;
		
		object_data = ObjectBox->item[i];

		if (strings_equal(name, object_data->name)) {
			
			return i;
		}
	}

	die("*** CORE ERROR *** invalid object: %s", name);

	/* NEVER REACHED */
	return OBJECT_INDEX_NIL;
}



/*
 * allocs object data
 */
static void * object_data_new(void)
{
	OBJECT_DATA *object_data;
	
	object_data = checked_malloc(sizeof *object_data);
	
	*object_data = ObjectDataTemplate;

	return object_data;
}



/*
 * frees object data
 */
static void object_data_free(void *data)
{

	free(data);
}

