/*
 * Copyright (C) 2002-2007 The Warp Rogue Team
 * Part of the Warp Rogue Project
 *
 * This software is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License.
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.
 *
 * See the license.txt file for more details.
 */

/*
 * Module Name: Talk
 * Description: -
 */

#include <stdarg.h>

#define Uses_Ui
#define Uses_Character
#define Uses_Perks
#define Uses_Util
#define Uses_Economy
#define Uses_Faction
#define Uses_Party
#define Uses_ProgramManager
#define Uses_Inventory
#include "mheader.h"

#include "talk.h"


#define MAX_GENERIC_DIALOG_OPTIONS	6


typedef enum {

	GDO_NIL = -1,
	
	GDO_RECRUIT,
	GDO_DISMISS,
	GDO_BUY,
	GDO_SELL,
	GDO_BUY_ROGUE,
	GDO_BYE,

	GDO_OUT_OF_BOUNDS

} GENERIC_DIALOG_OPTION;



static void		dialog_say(const char *, ...);

static void		dialog_add_option(const char *);
static const char *	dialog_choice(void);

static bool		generic_dialog(void);
static bool		recruiting_dialog(void);



static GENERIC_DIALOG_OPTION	string_to_generic_dialog_option(const char *);



static DIALOG_STATE	DialogState;

static const char *	GenericDialogOption[MAX_GENERIC_DIALOG_OPTIONS] = {

	"Join my party!",
	"Leave the party!",
	"I want to buy something.",
	"I want to sell something.",
	"I want to buy something special..",
	"Bye."

};




/*
 * starts a dialog
 */
void dialog(CHARACTER *active_character, CHARACTER *passive_character)
{

	DialogState.active_character = active_character;
	DialogState.passive_character = passive_character;

	if (character_has_flag(active_character, CF_CANNOT_SPEAK)) {

		message(SID_GAME, "You cannot speak.");

		return;
	}

	if (hostility_between(active_character, passive_character) ||
		character_has_flag(passive_character, CF_CANNOT_SPEAK)) {

		message(SID_GAME, "No reply.");

		return;
	}

	while (generic_dialog()) {
		/* REPEAT */;
	}
}



/*
 * makes the passive character say something
 */
static void dialog_say(const char *fmt, ...)
{
	char str[STRING_BUFFER_SIZE];
	va_list vl;

	va_start(vl, fmt);
	vsprintf(str, fmt, vl);
	va_end(vl);

	strcpy(DialogState.text, str);

	DialogState.options = list_new();
}



/*
 * adds a dialog option
 */
static void dialog_add_option(const char *str)
{

	list_add(DialogState.options, (char *)str);
}



/*
 * adds generic dialogs options
 */
static void dialog_add_generic_options(void)
{
	CHARACTER *character;

	character = DialogState.passive_character;

	if (character->perk[PK_TRADING]) {

		dialog_add_option(GenericDialogOption[GDO_BUY]);
		dialog_add_option(GenericDialogOption[GDO_SELL]);
	}
	
	if (character->perk[PK_ROGUE_TRADING]) {

		dialog_add_option(GenericDialogOption[GDO_BUY_ROGUE]);
	}

	if (character == player_character() ||
		character == DialogState.active_character) {

		/* recruiting/dismissing not possible */

	} else if (character->party == PARTY_PLAYER) {

		dialog_add_option(GenericDialogOption[GDO_DISMISS]);

	} else {

		dialog_add_option(GenericDialogOption[GDO_RECRUIT]);
	}

	dialog_add_option(GenericDialogOption[GDO_BYE]);
}



/*
 * lets the active character choose a dialog option
 */
static const char * dialog_choice(void)
{
	const char *chosen_option;

	DialogState.menu = menu_create(DialogState.options);

	command_bar_set(3, CM_UP, CM_DOWN, CM_OK);
	render_dialog_screen(&DialogState);
	update_screen();

	dialog_screen_menu(DialogState.menu);

	chosen_option = (const char *)list_data_at(DialogState.menu->items,
		DialogState.menu->highlighted);

	menu_destroy(DialogState.menu);

	return chosen_option;
}



/*
 * generic dialog
 */
static bool generic_dialog(void)
{
	const char *chosen_option;

	dialog_say(" ");

	dialog_add_generic_options();

	chosen_option = dialog_choice();

	switch (string_to_generic_dialog_option(chosen_option)) {

	case GDO_RECRUIT:
		return recruiting_dialog();

	case GDO_DISMISS:
		party_leave(DialogState.passive_character);
		return false;

	case GDO_BUY:
		buy_screen(DialogState.active_character, PK_TRADING);
		break;

	case GDO_SELL:
		sell_screen(DialogState.active_character);
		break;
	
	case GDO_BUY_ROGUE:
		buy_screen(DialogState.active_character, PK_ROGUE_TRADING);
		break;

	case GDO_BYE:
		return false;

	default:
		die("*** CORE ERROR *** invalid dialog option");
	}

	return true;
}



/*
 * recruiting dialog
 */
static bool recruiting_dialog(void)
{
	CHARACTER *recruiter;
	CHARACTER *recruit;
	const char *chosen_option;
	POWER_RATING party_power;
	MONEY price;

	recruiter = DialogState.active_character;
	recruit = DialogState.passive_character;

	if (!recruit->recruitable) {

		dialog_say("No, I am not interested in joining anyone.");

		dialog_add_option("...");
		dialog_choice();

		return true;
	}

	party_power = party_power_rating();

	if (recruiter->perk[PK_PERSUADE]) {

		party_power = percent(PERSUADE_POWER,
			party_power
		);
	}

	if (party_power < character_power_rating(recruit)) {

		dialog_say("No, your party looks weak.");

		dialog_add_option("...");
		dialog_choice();

		return true;
	}

	if (party_n_members() == MAX_PARTY_SIZE) {

		dialog_say("I think your party is big enough already.");
		dialog_add_option("...");
		dialog_choice();

		return true;
	}

	price = recruiting_cost(recruit);

	if (recruiter->perk[PK_PERSUADE]) {

		price = percent(PERSUADE_PRICE, price);
	}

	if (price == 0) {

		party_join(recruit);

		return false;
	}

	dialog_say("I demand %d credits.", price);
	dialog_add_option("Ok.");
	dialog_add_option("No way.");

	chosen_option = dialog_choice();

	if (strings_equal(chosen_option, "No way.")) {

		return true;
	}

	if (inventory_money(recruiter) < price) {

		dialog_say("So where are the credits?");
		dialog_add_option("...");
		dialog_choice();

		return true;
	}

	inventory_money_reduce(recruiter, price);

	party_join(recruit);

	return false;
}



/*
 * string -> generic dialog option
 */
static GENERIC_DIALOG_OPTION string_to_generic_dialog_option(const char *str)
{
	GENERIC_DIALOG_OPTION i;

	for (i = 0; i < MAX_OBJECT_ATTRIBUTES; i++) {

		if (strings_equal(str, GenericDialogOption[i])) {

			return i;
		}
	}

	die("*** CORE ERROR *** invalid generic dialog option (%s)",
		str
	);

	/* NEVER REACHED */
	return GDO_NIL;
}


