/*
 * Copyright (C) 2002-2007 The Warp Rogue Team
 * Part of the Warp Rogue Project
 *
 * This software is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License.
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY.
 *
 * See the license.txt file for more details.
 */

/*
 * Module Name: World
 * Description: -
 */

#define Uses_Ui
#define Uses_Util
#define Uses_Area
#define Uses_AreaTransition
#define Uses_ProgramManager
#define Uses_LoadSave
#include "mheader.h"

#include "world.h"


/*
 * tile box buffer size
 */
#define TILE_BOX_BUFFER_SIZE	16


/*
 * number of world map tiles
 */
typedef int			N_MAP_TILES;




static void *			tile_data_new(void);
static void			tile_data_free(void *);





/*
 * the map tile box
 */
static BOX *			TileBox = NULL;


/*
 * map tile data template (default values)
 */
static const MAP_TILE_DATA	TileDataTemplate = {
	
	/* NAME */ "", /* SYMBOL */ 'E', /* COLOR */ C_RED,
	/* DUNGEON_FILE */ ""

};



/*
 * the world map
 */
static MAP_TILE WorldMap[WORLD_MAX_Y + 1][WORLD_MAX_X +1];



/*
 * World module init
 */
void world_init(void)
{

	TileBox = box_new(tile_data_new, tile_data_free, 
		TILE_BOX_BUFFER_SIZE
	);
}


/*
 * World module clean up
 */
void world_clean_up(void)
{

	if (TileBox != NULL) {

		box_free(TileBox);
	}
}



/*
 * adds a map tile to the tile box
 */
MAP_TILE_DATA * world_tile_box_new_tile(void)
{

	return box_new_item(TileBox);
}



/*
 * optimizes the tile box
 */
void world_tile_box_optimize(void)
{

	box_optimize(TileBox);
}



/*
 * returns a pointer to the requested world map tile
 */
MAP_TILE * world_map_tile(const WORLD_POINT *p) 
{

	return &WorldMap[p->y][p->x];
}



/*
 * returns a pointer to the data of the passed tile
 */
MAP_TILE_DATA * world_map_tile_data(const MAP_TILE *tile)
{

	return TileBox->item[*tile];
}



/*
 * returns true if the coordinates of both points are equal
 */
bool world_points_equal(const WORLD_POINT *p1, const WORLD_POINT *p2)
{

	if (p1->z == p2->z && p1->y == p2->y &&
		p1->x == p2->x) {

		return true;
	}

	return false;
}



/*
 * moves a world point one step in the passed direction
 */
void move_world_point(WORLD_POINT *world_point, DIRECTION direction)
{

	world_point->x += direction_modifier(COORD_X, direction);
	world_point->y += direction_modifier(COORD_Y, direction);	
}



/*
 * returns true if the passed point is out of bounds
 */
bool out_of_world_bounds(const WORLD_POINT *p)
{

	if (p->z < 0 || p->y < 0 || p->x < 0 ||
		p->z > WORLD_MAX_Z || p->y > WORLD_MAX_Y ||
		p->x > WORLD_MAX_X) {

		return true;
	}

	return false;
}



/*
 * world map screen
 */
void world_map_screen(const CHARACTER *character)
{
	COMMAND entered_command;
	DIRECTION direction;
	WORLD_POINT current_position;
	
	current_position = active_area()->location;
	
	command_bar_clear();
	command_bar_add_move_commands();

	while (true) {
		WORLD_POINT previous_position;

		render_world_map_screen(&current_position, character);

		update_screen();

		entered_command = command_bar_get_command();
	
		move_command_to_direction(&direction, entered_command);

		previous_position = current_position;

		move_world_point(&current_position, direction);
		
		if (out_of_world_bounds(&current_position)) {
			
			current_position = previous_position;
			
			continue;
		}
		
		if (area_exists(&current_position)) {
			
			move_party_to_area(&current_position, direction);
			
			break;
		}
	}
}



/*
 * symbol -> world map tile
 */
MAP_TILE symbol_to_world_map_tile(SYMBOL symbol)
{
	MAP_TILE tile;

	for (tile = 0; tile < TileBox->current_size; tile++) {
		const MAP_TILE_DATA *tile_data;
		
		tile_data = TileBox->item[tile];

		if (tile_data->symbol == symbol) {

			return tile;
		}
	}

	die("*** CORE ERROR *** invalid world map symbol: %c", symbol);

	/* NEVER REACHED */
	return 0;
}



/*
 * allocs tile data
 */
static void * tile_data_new(void)
{
	MAP_TILE_DATA *tile_data;
	
	tile_data = checked_malloc(sizeof *tile_data);
	
	*tile_data = TileDataTemplate;

	return tile_data;
}



/*
 * frees tile data
 */
static void tile_data_free(void *data)
{

	free(data);
}
